import unittest
from unittest.mock import patch

from killstats.api.account_manager import AccountManager

MODULE_PATH = "killstats.api.account_manager"


class TestAccountManager(unittest.TestCase):
    def setUp(self):
        self.account_manager = AccountManager()
        self.account_manager.corporations = {1, 2, 3}
        self.account_manager.alliances = {10, 20, 30}

    @patch(MODULE_PATH + ".EveCharacter")
    def test_process_character_in_corporation(self, MockEveCharacter):
        char = MockEveCharacter()
        char.corporation_id = 1
        char.alliance_id = 100
        char.character_ownership.user.profile.main_character = char
        characters = {}
        chars_list = set()
        missing_chars = set()

        self.account_manager._process_character(
            char, characters, chars_list, missing_chars
        )

        self.assertIn(char.character_id, chars_list)
        self.assertIn(char, characters[char.character_id]["alts"])

    @patch(MODULE_PATH + ".EveCharacter")
    def test_process_character_in_alliance(self, MockEveCharacter):
        char = MockEveCharacter()
        char.corporation_id = 100
        char.alliance_id = 10
        char.character_ownership.user.profile.main_character = char
        characters = {}
        chars_list = set()
        missing_chars = set()

        self.account_manager._process_character(
            char, characters, chars_list, missing_chars
        )

        self.assertIn(char.character_id, chars_list)
        self.assertIn(char, characters[char.character_id]["alts"])

    @patch(MODULE_PATH + ".EveCharacter")
    def test_process_character_not_in_corporation_or_alliance(self, MockEveCharacter):
        char = MockEveCharacter()
        char.corporation_id = 100
        char.alliance_id = 100
        char.character_ownership.user.profile.main_character = char
        characters = {}
        chars_list = set()
        missing_chars = set()

        self.account_manager._process_character(
            char, characters, chars_list, missing_chars
        )

        self.assertNotIn(char.character_id, chars_list)
        self.assertNotIn(char, characters.get(char.character_id, {}).get("alts", []))



# Django
from django.db import models
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity, EveType

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# AA Killstats
from killstats.hooks import get_extension_logger
from killstats.managers.killboard_manager import EveKillmailManager

logger = get_extension_logger(__name__)


class Killmail(models.Model):
    killmail_id = models.PositiveBigIntegerField(primary_key=True)
    killmail_date = models.DateTimeField(null=True, blank=True)
    victim = models.ForeignKey(EveEntity, on_delete=models.CASCADE, null=True)
    victim_ship = models.ForeignKey(EveType, on_delete=models.CASCADE, null=True)
    victim_corporation_id = models.PositiveBigIntegerField()
    victim_alliance_id = models.PositiveBigIntegerField(null=True, blank=True)
    hash = models.CharField(max_length=255, unique=True)
    # Value Infos
    victim_total_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_fitted_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_destroyed_value = models.PositiveBigIntegerField(null=True, blank=True)
    victim_dropped_value = models.PositiveBigIntegerField(null=True, blank=True)
    # Location Infos
    victim_region_id = models.PositiveBigIntegerField(null=True, blank=True)
    victim_solar_system_id = models.PositiveBigIntegerField(null=True, blank=True)
    victim_position_x = models.FloatField(null=True, blank=True)
    victim_position_y = models.FloatField(null=True, blank=True)
    victim_position_z = models.FloatField(null=True, blank=True)

    objects = EveKillmailManager()

    def __str__(self):
        return f"Killmail {self.killmail_id}"

    def get_or_unknown_victim_name(self):
        """Return the victim name or Unknown."""
        return self.victim.name if self.victim else _("Unknown")

    def get_or_unknown_victim_ship_name(self):
        """Return the victim ship name or Unknown."""
        return self.victim_ship.name if self.victim_ship else _("Unknown")

    def evaluate_victim_id(self):
        """Return the victim ID."""
        if self.victim is not None:
            return self.victim.id
        if self.victim_corporation_id is not None:
            return self.victim_corporation_id
        if self.victim_alliance_id is not None:
            return self.victim_alliance_id
        return 0

    def evaluate_zkb_link(self):
        zkb = f"https://zkillboard.com/character/{self.victim.id}/"
        if self.victim.category == "corporation":
            zkb = f"https://zkillboard.com/corporation/{self.victim_corporation_id}/"
        if self.victim.category == "alliance":
            zkb = f"https://zkillboard.com/alliance/{self.victim_alliance_id}/"
        return zkb

    def evaluate_portrait(self):
        portrait = eveimageserver.character_portrait_url(self.victim.id, 256)
        if self.victim.category == "corporation":
            portrait = eveimageserver.corporation_logo_url(
                self.victim_corporation_id, 256
            )
        if self.victim.category == "alliance":
            portrait = eveimageserver.alliance_logo_url(self.victim_alliance_id, 256)
        return portrait

    def get_ship_image_url(self):
        if self.victim_ship is not None:
            return eveimageserver.type_render_url(self.victim_ship.id, 32)
        return ""

    def get_image_url(self):
        if self.victim is not None:
            return eveimageserver._eve_entity_image_url(
                self.victim.category, self.victim.id, 32
            )
        return ""

    def is_corp(self, corps: list):
        """Return True if the victim corporation is in the list of corporations."""
        attackers = Attacker.objects.filter(killmail=self).values_list(
            "corporation_id", flat=True
        )
        return self.victim_corporation_id in corps or any(
            attacker in corps for attacker in attackers
        )

    def is_alliance(self, alliances: list):
        """Return True if the victim alliance is in the list of alliances."""
        attackers = Attacker.objects.filter(killmail=self).values_list(
            "alliance_id", flat=True
        )
        return self.victim_alliance_id in alliances or any(
            attacker in alliances for attacker in attackers
        )

    def is_structure(self):
        """Return True if the victim is a structure."""
        return self.victim_ship.eve_group.eve_category_id == 65

    def is_mobile(self):
        """Return True if the victim is a mobile structure."""
        return self.victim_ship.eve_group.eve_category_id == 22

    def is_capsule(self):
        """Return True if the victim is a capsule."""
        return self.victim_ship.eve_group.id == 29

    def get_month(self, month):
        """Get all killmails of a specific month."""
        return self.killmail_date.month == month

    def threshold(self, threshold: int) -> bool:
        """Return True if the total value of the killmail is above the threshold."""
        return self.victim_total_value > threshold

    class Meta:
        default_permissions = ()


class Attacker(models.Model):
    killmail = models.ForeignKey(
        Killmail, on_delete=models.CASCADE, related_name="attacker_killmail"
    )
    character = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        related_name="attacker_character",
        null=True,
        blank=True,
    )
    corporation = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        related_name="attacker_corp",
        null=True,
        blank=True,
    )
    alliance = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        related_name="attacker_alliance",
        null=True,
        blank=True,
    )
    ship = models.ForeignKey(
        EveType,
        on_delete=models.CASCADE,
        related_name="attacker_ship",
        null=True,
        blank=True,
    )
    damage_done = models.IntegerField(null=True, blank=True)
    final_blow = models.BooleanField(null=True, blank=True)
    weapon_type_id = models.PositiveBigIntegerField(null=True, blank=True)
    security_status = models.FloatField(null=True, blank=True)

    def get_or_unknown_ship_name(self):
        """Return the ship name or Unknown."""
        return self.ship.name if self.ship else _("Unknown")

    def evaluate_attacker(self) -> tuple:
        """Return the attacker ID and Name."""
        if self.character is not None:
            return self.character.id, self.character.name
        if self.corporation is not None:
            return self.corporation.id, self.corporation.name
        if self.alliance is not None:
            return self.alliance.id, self.alliance.name
        return 0, _("Unknown")

    class Meta:
        default_permissions = ()



from unittest.mock import MagicMock, patch

from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from killstats.api.killboard.killboard_helper import (
    get_killstats_halls,
    get_killstats_stats,
)
from killstats.tests.testdata.load_allianceauth import load_allianceauth
from killstats.tests.testdata.load_killstats import load_killstats_all

MODULE_PATH = "killstats.api.killboard.killboard_helper"


class ManageApiCorporationEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_killstats_all()

        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=["killstats.basic_access", "killstats.admin_access"],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "killstats.basic_access",
            ],
        )

    def test_get_killstats_halls_filter_corporations(self):
        request = MagicMock()
        request.user = self.user
        month = 1
        year = 2023
        entity_id = 0
        entity_type = "corporation"

        with patch(MODULE_PATH + ".get_corporations", return_value=[2001, 10000001]):
            halls = get_killstats_halls(request, month, year, entity_id, entity_type)
            self.assertIsNotNone(halls)
            self.assertIn("shame", halls[0])
            self.assertIn("fame", halls[0])
            self.assertNotIn(10000001, halls[0]["shame"])
            self.assertNotIn(10000001, halls[0]["fame"])

    def test_get_killstats_halls_filter_alliances(self):
        request = MagicMock()
        request.user = self.user
        month = 1
        year = 2023
        entity_id = 0
        entity_type = "alliance"

        with patch(MODULE_PATH + ".get_alliances", return_value=[3001, 10000001]):
            stats = get_killstats_stats(request, month, year, entity_id, entity_type)

            self.assertIsNotNone(stats)
            self.assertIn("stats", stats[0])



"""Managers for killboard."""

# Standard Library
from typing import Any

# Django
from django.db import models, transaction
from eveuniverse.models import EveEntity, EveType

from killstats.decorators import log_timing

# AA Voices of War
from killstats.hooks import get_extension_logger
from killstats.managers.killmail_core import KillmailManager

logger = get_extension_logger(__name__)


class KillmailQueryCore(models.QuerySet):
    def filter_entities(self, entities):
        """Filter Kills and Losses from Entities List (Corporations or Alliances)."""
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Attacker

        kms = []

        # Get all Killmail IDs
        km_ids = self.values_list("killmail_id", flat=True)

        # Get all Attackers from Entities
        kms_data = Attacker.objects.filter(
            models.Q(corporation_id__in=entities)
            | models.Q(alliance_id__in=entities)
            | models.Q(character_id__in=entities),
            killmail_id__in=km_ids,
        ).values_list("killmail_id", "corporation_id", "alliance_id")

        for killmail_id, corporation_id, alliance_id in kms_data:
            if corporation_id in entities or alliance_id in entities:
                kms.append(killmail_id)

        # Include Victim Kills
        victim_kms_ids = self.filter(
            models.Q(victim_id__in=entities)
            | models.Q(victim_corporation_id__in=entities)
            | models.Q(victim_alliance_id__in=entities)
        ).values_list("killmail_id", flat=True)

        for killmail_id in victim_kms_ids:
            kms.append(killmail_id)

        return self.filter(killmail_id__in=kms)

    def filter_entities_kills(self, entities):
        """Filter Kills from Entities List (Corporations or Alliances)."""
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Attacker

        kms = []

        # Sammle alle Angreifer-Daten
        km_ids = self.values_list("killmail_id", flat=True)

        kms_data = Attacker.objects.filter(
            models.Q(corporation_id__in=entities)
            | models.Q(alliance_id__in=entities)
            | models.Q(character_id__in=entities),
            killmail_id__in=km_ids,
        ).values_list("killmail_id", "corporation_id", "alliance_id")

        for killmail_id, corporation_id, alliance_id in kms_data:
            if corporation_id in entities or alliance_id in entities:
                kms.append(killmail_id)

        return self.filter(killmail_id__in=kms)

    def filter_entities_losses(self, entities):
        """Filter Losses from Entities List (Corporations or Alliances)."""
        kms = []

        victim_kms = self.filter(
            models.Q(victim_id__in=entities)
            | models.Q(victim_corporation_id__in=entities)
            | models.Q(victim_alliance_id__in=entities)
        )

        for killmail in victim_kms:
            kms.append(killmail.killmail_id)

        return self.filter(killmail_id__in=kms)

    def filter_structure(self, exclude=False):
        """Filter or Exclude Structure Kills."""
        if exclude:
            return self.exclude(victim_ship__eve_group__eve_category_id=65)
        return self.filter(victim_ship__eve_group__eve_category_id=65)

    def filter_threshold(self, threshold: int):
        """Filter Killmails are in Threshold."""
        return self.filter(victim_total_value__gt=threshold)


class KillmailQueryMining(KillmailQueryCore):
    def filter_barge(self):
        """Filter Mining Barge."""
        return self.filter(victim_ship__eve_group_id=463)

    def filter_exhumer(self):
        """Filter Exhumer."""
        return self.filter(victim_ship__eve_group_id=543)

    def filter_indu_command_ship(self):
        """Filter Industrial Command Ship."""
        return self.filter(victim_ship__eve_group_id=941)

    def filter_capital_indu_ship(self):
        """Filter Capital Industrial Ship."""
        return self.filter(victim_ship__eve_group_id=883)


class KillmailQueryStats(KillmailQueryMining):
    @log_timing(logger)
    def _get_top_ship(self, entities, km_ids):
        """Get the top ship for the given entities."""
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Attacker

        topship = (
            Attacker.objects.filter(
                models.Q(corporation_id__in=entities)
                | models.Q(alliance_id__in=entities)
                | models.Q(character_id__in=entities),
                killmail_id__in=km_ids,
            )
            .values("ship__id")
            .annotate(count=models.Count("ship__id"))
            .order_by("-count")
        ).first()

        if topship:
            top_ship = EveType.objects.get(id=topship["ship__id"])
            top_ship.ship_count = topship["count"]
            return top_ship
        return None

    @log_timing(logger)
    def _get_worst_ship(self, entities, km_ids):
        """Get the worst ship for the given entities."""
        losses = self.filter(
            models.Q(victim_id__in=entities)
            | models.Q(victim_corporation_id__in=entities)
            | models.Q(victim_alliance_id__in=entities),
            killmail_id__in=km_ids,
        ).exclude(victim_ship__eve_group_id=29)

        worst_ship = (
            losses.values("victim_ship_id")
            .annotate(kill_count=models.Count("victim_ship_id"))
            .order_by("-kill_count", "victim_ship__name")
        ).first()

        if worst_ship:
            ship_id = worst_ship["victim_ship_id"]
            kill_count = worst_ship["kill_count"]

            # Get the worst ship and add the kill count
            worst_ship = EveType.objects.get(id=ship_id)
            worst_ship.ship_count = kill_count
            return worst_ship

        return None

    @log_timing(logger)
    def _get_top_victim(self, entities, km_ids):
        """Get the top victim for the given entities."""
        victim = (
            self.filter(
                models.Q(victim_id__in=entities)
                | models.Q(victim_corporation_id__in=entities)
                | models.Q(victim_alliance_id__in=entities),
                killmail_id__in=km_ids,
            )
            .values("victim_id")
            .annotate(kill_count=models.Count("victim_id"))
            .order_by("-kill_count", "victim__name")
        ).first()

        if victim:
            top_victim = EveEntity.objects.get(id=victim["victim_id"])
            top_victim.kill_count = victim["kill_count"]
            return top_victim
        return None

    @log_timing(logger)
    def _get_top_killer(self, entities, km_ids):
        """Get the top killer for the given entities."""
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Attacker

        attacker = (
            Attacker.objects.filter(
                models.Q(corporation_id__in=entities)
                | models.Q(alliance_id__in=entities)
                | models.Q(character_id__in=entities),
                killmail_id__in=km_ids,
            )
            .values("character_id")
            .annotate(kill_count=models.Count("killmail_id"))
            .order_by("-kill_count", "character__name")
        ).first()

        if attacker:
            top_attacker = EveEntity.objects.get(id=attacker["character_id"])
            top_attacker.kill_count = attacker["kill_count"]
            return top_attacker
        return None

    @log_timing(logger)
    def get_killboard_stats(self, entities, date):
        """Get all killboard stats for the given entities."""
        stats = {}

        month, year = date.month, date.year

        # Common queries
        losses = self.filter_entities_losses(entities)
        kills = self.filter_entities_kills(entities)

        # Get the highest loss
        highest_loss = (
            losses.filter(killmail_date__year=year, killmail_date__month=month)
            .annotate(total_value=models.F("victim_total_value"))
            .order_by("-total_value", "-victim_fitted_value")
            .first()
        )
        stats["highest_loss"] = highest_loss

        # Get the highest kill
        highest_kill = (
            kills.filter(killmail_date__year=year, killmail_date__month=month)
            .annotate(total_value=models.F("victim_total_value"))
            .order_by("-total_value", "-victim_fitted_value")
            .first()
        )
        stats["highest_kill"] = highest_kill

        # Get Killmail IDs for the given year
        km_ids_year = kills.filter(killmail_date__year=year).values_list(
            "killmail_id", flat=True
        )
        # Get Killmail IDs for the given month
        km_ids = km_ids_year.filter(
            killmail_date__year=year,
            killmail_date__month=month,
        ).values_list("killmail_id", flat=True)
        # Get Loss Killmail IDs for the given year
        km_ids_loss_year = losses.filter(killmail_date__year=year).values_list(
            "killmail_id", flat=True
        )
        # Get Loss Killmail IDs for the given month
        km_ids_loss = km_ids_loss_year.filter(
            killmail_date__year=year,
            killmail_date__month=month,
        ).values_list("killmail_id", flat=True)

        # Get the worst ship
        stats["worst_ship"] = self._get_worst_ship(entities, km_ids_loss)

        stats["top_ship"] = self._get_top_ship(entities, km_ids)

        # Get the top killer Month
        stats["top_killer"] = self._get_top_killer(entities, km_ids)

        # Get the top killer Year
        stats["alltime_killer"] = self._get_top_killer(entities, km_ids_year)

        # Get the top victim Month
        stats["top_loss"] = self._get_top_victim(entities, km_ids_loss)

        # Get the top victim Year
        stats["alltime_loss"] = self._get_top_victim(entities, km_ids_loss_year)

        return stats


class KillmailQuerySet(KillmailQueryStats):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all Squads for superuser %s.", user)
            return self

        if user.has_perm("killstats.admin_access"):
            logger.debug("Returning all Killboards for Admin %s.", user)
            return self

        return self.none()


class KillmailBaseManager(models.Manager):
    def get_queryset(self):
        return KillmailQuerySet(self.model, using=self._db)

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def create_from_killmail(self, killmail: KillmailManager):
        """create a new EveKillmail from a Killmail object and returns it"""
        # AA Killstats
        # pylint: disable=import-outside-toplevel
        from killstats.models.killboard import Killmail

        with transaction.atomic():
            attackers_list = []

            corporation_id = killmail.victim.corporation_id
            region_id = killmail.get_region_id(killmail.solar_system_id)
            victim_ship = killmail.get_ship_name(killmail.victim.ship_type_id)
            victim = None

            if killmail.victim.character_id:
                victim = killmail.get_entity_name(killmail.victim.character_id)
            elif killmail.victim.alliance_id:
                victim = killmail.get_entity_name(killmail.victim.alliance_id)
            elif killmail.victim.corporation_id:
                victim = killmail.get_entity_name(killmail.victim.corporation_id)

            for attacker in killmail.attackers:
                if attacker.character_id:
                    attackers_list.append(attacker.character_id)

            if attackers_list:
                try:
                    unique_attackers = list(set(attackers_list))
                    killmail.create_names_bulk(eve_ids=unique_attackers)
                # pylint: disable=broad-exception-caught
                except Exception as e:
                    logger.debug("Error on Create Names: %s", e, exc_info=True)

            km = Killmail.objects.create(
                killmail_id=killmail.id,
                killmail_date=killmail.time,
                victim=victim,
                victim_ship=victim_ship,
                victim_corporation_id=corporation_id,
                victim_alliance_id=killmail.victim.alliance_id,
                hash=killmail.zkb.hash,
                victim_total_value=killmail.zkb.total_value,
                victim_fitted_value=killmail.zkb.fitted_value,
                victim_destroyed_value=killmail.zkb.destroyed_value,
                victim_dropped_value=killmail.zkb.dropped_value,
                victim_region_id=region_id,
                victim_solar_system_id=killmail.solar_system_id,
                victim_position_x=killmail.position.x,
                victim_position_y=killmail.position.y,
                victim_position_z=killmail.position.z,
            )

            killmail.create_attackers(km, killmail)

        return km

    def update_or_create_from_killmail(
        self, killmail: KillmailManager
    ) -> tuple[Any, bool]:
        """Update or create new EveKillmail from a Killmail object."""
        with transaction.atomic():
            try:
                self.get(killmail_id=killmail.id).delete()
                created = False
            except self.model.DoesNotExist:
                created = True
            obj = self.create_from_killmail(killmail)
        return obj, created


EveKillmailManager = KillmailBaseManager.from_queryset(KillmailQuerySet)
